#!/usr/bin/python

# Generates an ASCII representation of a ToH problem after k applied moves 
# following the Frame-Stewart algorithm.

# Written by Alex Westlund and Shaden Smith under Mirek Truszczynski
# Fall 2010

# USAGE: ./stateGenerator numPegs numDisks numMoves

import sys
import string

PEGS = 0
DISKS = 0

end = 0
count = 1

solLength = []	# 2D List
kTable = []

pegs = [] # stacks representing pegs

def prepTable():
	global solLength
	global kTable

	# Fill in tables with -1. We work with PEGS-2 because 1 and 2 pegs 
	# shouldn't be calculated.
	for i in range(PEGS-2):
		solLength.append([-1]*(DISKS+1))
		kTable.append([-1]*(DISKS+1))
	
	# 3 Peg solutions are easy to calculate.
	# Solutions are 2^n - 1 and k values are always DISKS-1 because of the
	# nature of the recursive formula involved in solving the 3-peg game.
	for i in range(DISKS+1):
		solLength[0][i] = (2**i)-1
		kTable[0][i] = i-1
	
	# for >3 pegs the first few disk values are all the same.
	for i in range(1,PEGS-2):
		solLength[i][0] = 0
		solLength[i][1] = 1
		solLength[i][2] = 3
		solLength[i][3] = 5
		
		kTable[i][2] = 1
		kTable[i][3] = 1


# Return optimal solution length with n pegs on r disks.
# Solves recurrence relation of the Frame-Stewart algorithm:
# 2T(k,r) + T(n-k,r-1)
# where k is some number >= 1
def T(n,r):
	global solLength
	global kTable

	# memoization
	if solLength[r][n] > -1:
		return solLength[r][n]
	
	maxK = n-1
	minimum = -1
	for k in range(1,maxK):
		p1 = 2*T(k,r)
		p2 = T(n-k,r-1)
		curr = p1 + p2

		if curr < minimum or minimum == -1:
			minimum = curr
			kTable[r][n] = k

	solLength[r][n] = minimum
	return minimum

# retrieve optimized k value for r pegs and n disks
def getK(n, r):
	global kTable
	global solLength

	r -= 3

	if solLength[r][n] == -1:
		T(n,r)
	
	return kTable[r][n]

# helper function for T. T requires n and r to be modified since solLength only
# actually has PEGS-2 columns.
def numMoves(n, r):
	r -= 3
	return T(n, r)

# Solve count "optimal" moves.
def solve(n, r, src, dest, game):
	global end
	global count
	# Base case, only one disk remaining on peg, so move
	if n == 1:
		move(game[src], game[dest])
		
		count += 1 
		if count == end+1 or end < 0: 	# -1 prints all moves, if we've reached
			printState(game)			# end, exit
			if count == end+1:
				sys.exit(0)
		return
	# move top k disks to next free peg with no move restrictions
	k = getK(n, r)
	# nextFree determines where to move auxiliary disks. Start at the beginning
	# of pegs and select first valid move that isn't a previous destination
	nextFree = -1
	i = 0

	while nextFree == -1:
		# if it's a valid move
		if (len(game[i]) == 0 or game[i][-1] < game[src][-1]) and i != dest:
			nextFree = i
		i += 1

	# move top k disks to next free peg
	solve(k, r, src, nextFree, game)

	# now move the remaining n-k disks to destination peg without disturbing
	# nextFree
	solve(n-k, r-1, src, dest, game)

	# now move original k disks from nextFree to destination
	solve(k, r, nextFree, dest, game)

# Moves top disk A->B
# returns true if valid move, false if failed.
# Validity Test: Larger numbers must always be on top.
def move(a, b):
	# if a is empty, illegal
	if len(a) == 0:
		return False
	# either condition would constitute a valid move
	if len(b) == 0 or a[-1] > b[-1]:
		b.append(a[-1])
		a.pop()
		return True
	
	# failed
	return False

def printState(game):#data):
	char = 0
	#game = copy.deepcopy(data) # we don't want to alter the peg data
	for i in range(len(game)):
		currPeg = string.uppercase[char]
		#game[i].reverse()
	
		p = ""
		for x in game[i]:
			p += str(x) + " "
		
		char += 1 		# increment peg name
		print "%s: %s" % (currPeg, p)

	print "\n"

def main():
	global PEGS
	global DISKS
	global pegs
	global end
	global count

	# Grab input and validate
	if len(sys.argv) < 4:
		print "Usage: %s pegs disks numMoves" % sys.argv[0]
		return
	
	PEGS = int(sys.argv[1])
	DISKS = int(sys.argv[2])	# + 1 due to counting 0 disks in the array
	end = int(sys.argv[3])	# state after count moves
	
	if PEGS < 3:
		print "3 Peg minimum."
		return
	if DISKS < 4:
		print "4 Disk minimum."
		return

	prepTable()

	# header output
	print "Pegs: %d\nDisks: %d\nMoves: %d\n" % (PEGS, DISKS, numMoves(DISKS, PEGS))

	# fill the first peg
	for i in range(PEGS):
		pegs.append([])
	for i in range(1, DISKS+1):
		pegs[0].append(i)

	# 0 means don't print a state, just exit
	if end == 0:
		printState(pegs)
		return
	# 0 means don't print a state, just exit
	if end == 0:
		printState(pegs)
		return

	# generate state information
	solve(DISKS, PEGS, 0, PEGS-1, pegs) # n, r, src, dest, pegs

if __name__ == "__main__":
	main()
